package net.pingex.dcf.commands;

import net.pingex.dcf.commands.options.ICommandOption;
import sx.blah.discord.handle.impl.events.MessageReceivedEvent;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Set;

/**
 * Represents a single command.
 */
public abstract class Command implements ICommandExecutor
{
	/**
	 * Main name of the command
	 */
	private String name;

	/**
	 * Command can also be called using the following list of aliases
	 */
	private List<String> aliases;

	/**
	 * Description of the command.
	 */
	private String description;

	/**
	 * Is the command enabled ? Can it be invoked ?
	 */
	private boolean isEnabled;

	/**
	 * Command usage help
	 */
	private String usage;

	/**
	 * Contains all options for this command.
	 */
	private Set<ICommandOption> commandOptions;

	/**
	 * Basic constructor.
	 * @param name Name of the command
	 * @param aliases Aliases, if any
	 * @param description Description of the command
	 * @param isEnabled Is the command enabled ?
	 * @param usage Command usage help
	 * @param options Command options.
	 */
	public Command(String name, List<String> aliases, String description, boolean isEnabled, String usage, Set<ICommandOption> options)
	{
		this.name = name;
		this.aliases = aliases;
		this.description = description;
		this.isEnabled = isEnabled;
		this.usage = usage;
		this.commandOptions = options;
	}

	/**
	 * Contains Command default values
	 */
	public static class Defaults
	{
		public static final List<String> ALIASES = Collections.emptyList();
		public static final String DESCRIPTION = "No command description provided.";
		public static final boolean IS_ENABLED = true;
		public static final String USAGE = "No command usage provided.";
		public static final Set<ICommandOption> OPTIONS = Collections.emptySet();
	}

	/**
	 * Builder for a command
	 */
	public static class Builder
	{
		/**
		 * Main name of the command
		 */
		private String name;

		/**
		 * Command can also be called using the following list of aliases
		 */
		private List<String> aliases = Defaults.ALIASES;

		/**
		 * Description of the command.
		 */
		private String description = Defaults.DESCRIPTION;

		/**
		 * Is the command enabled ? Can it be invoked ?
		 */
		private boolean isEnabled = Defaults.IS_ENABLED;

		/**
		 * Command usage help
		 */
		private String usage = Defaults.USAGE;

		/**
		 * Command options
		 */
		private Set<ICommandOption> commandOptions = Defaults.OPTIONS;

		public Builder(String name)
		{
			this.name = name;
		}

		public Builder aliases(List<String> aliases)
		{
			this.aliases = aliases;
			return this;
		}

		/**
		 * Set aliases using varargs
		 * @param aliases Aliases varargs
		 */
		public Builder aliases(String... aliases)
		{
			this.aliases = Arrays.asList(aliases);
			return this;
		}

		public Builder description(String description)
		{
			this.description = description;
			return this;
		}

		public Builder enabled(boolean enabled)
		{
			isEnabled = enabled;
			return this;
		}

		public Builder usage(String usage)
		{
			this.usage = usage;
			return this;
		}

		public Builder options(Set<ICommandOption> options)
		{
			commandOptions = options;
			return this;
		}

		/**
		 * Build a new Command using a supplied executor.
		 * @param toExecute The body of the command.
		 * @return Built command.
		 */
		public Command build(ICommandExecutor toExecute)
		{
			return new Command(name, aliases, description, isEnabled, usage, commandOptions)
			{
				@Override
				public void execute(MessageReceivedEvent event, List<String> arguments) throws Throwable
				{
					toExecute.execute(event, arguments);
				}
			};
		}

		/**
		 * Build a new Command using a method and a target object (for use with annotated command)
		 * @param target The method which needs to be invoked
		 * @param invokable The matching object
		 * @return Built command.
		 */
		public Command build(Method target, Object invokable)
		{
			return new Command(name, aliases, description, isEnabled, usage, commandOptions)
			{
				@Override
				public void execute(MessageReceivedEvent event, List<String> arguments) throws Throwable
				{
					target.invoke(invokable, event, arguments);
				}
			};
		}
	}

	/**
	 * Get a builder for a new command.
	 * @param name The name of the future command
	 * @return A new manipulable builder to build the command.
	 */
	public static Builder builder(String name)
	{
		return new Builder(name);
	}

	public String getName()
	{
		return name;
	}

	public List<String> getAliases()
	{
		return Collections.unmodifiableList(aliases);
	}

	public String getDescription()
	{
		return description;
	}

	public boolean isEnabled()
	{
		return isEnabled;
	}

	public String getUsage()
	{
		return usage;
	}

	public Set<ICommandOption> getOptions()
	{
		return Collections.unmodifiableSet(commandOptions);
	}

	public void setEnabled(boolean enabled)
	{
		isEnabled = enabled;
	}
}
