package net.pingex.dcf.util;

import org.apache.commons.lang3.ClassUtils;
import sx.blah.discord.api.IDiscordClient;
import sx.blah.discord.handle.obj.IGuild;
import sx.blah.discord.handle.obj.IUser;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.regex.Pattern;

/**
 * Util class used to parse a String into primitive types.
 */
public class ArgumentParser
{
	/**
	 * Parses a String into a primitive variable.
	 * Choice between boolean, byte, short, int, long, float, double, String and their respective wrapper types.
	 * @param target Target primitive
	 * @param value Input value
	 * @return `null` if the value is null or the string content is null, or the parsed value.
	 * @throws IllegalArgumentException If the target primitive/class isn't part of the previous list.
	 * @throws NumberFormatException if the cast failed.
	 */
	public static Object parse(Class<?> target, String value) throws IllegalArgumentException
	{
		if(value         == null   || value.equalsIgnoreCase("null")) return null;
		if(Boolean.class == target || Boolean.TYPE == target) return Boolean.parseBoolean(value);
		if(Byte.class    == target || Byte.TYPE    == target) return Byte.parseByte(value);
		if(Short.class   == target || Short.TYPE   == target) return Short.parseShort(value);
		if(Integer.class == target || Integer.TYPE == target) return Integer.parseInt(value);
		if(Long.class    == target || Long.TYPE    == target) return Long.parseLong(value);
		if(Float.class   == target || Float.TYPE   == target) return Float.parseFloat(value);
		if(Double.class  == target || Double.TYPE  == target) return Double.parseDouble(value);
		if(String.class  == target)                           return value;
		throw new IllegalArgumentException("Unknown target type");
	}

	/**
	 * Parse the input arguments to a list of argument of various types.
	 * Use a primitive type for an argument that is required.
	 * Use a wrapper type to an argument that is optional, thereby the said argument can be {@code null}.
	 * Any extra argument not specified in {@code targetTypes} are ignored and will not appear in the output list.
	 *
	 * @param targetTypes A list of types that the {@code arguments} will be checked against.
	 * @param arguments Input arguments to check
	 * @return The parsed arguments whose types matches {@code targetTypes}
	 * @throws ParserException if the parsing failed.
	 */
	public static List<Object> parseAll(List<Class<?>> targetTypes, List<String> arguments) throws ParserException
	{
		if(targetTypes.size() == 0) throw new ParserException("Nothing to parse.");

		List<Object> outputList = new ArrayList<>();

		for(int i = 0; i < targetTypes.size(); i++)
		{
			try
			{
				Object parsingOutput = parse(targetTypes.get(i), arguments.get(i));
				if(parsingOutput == null)
				{
					if(ClassUtils.isPrimitiveWrapper(targetTypes.get(i)) || targetTypes.get(i) == String.class) outputList.add(null); // Allows null
					else throw new ParserException("Invalid arguments."); // Disallows null
				}
				else outputList.add(parsingOutput);
			}
			catch(IndexOutOfBoundsException e) // Not enough arguments supplied
			{
				if(ClassUtils.isPrimitiveWrapper(targetTypes.get(i)) || targetTypes.get(i) == String.class) outputList.add(null); // Allows null
				else throw new ParserException("Not enough arguments supplied."); // Disallows null
			}
			catch(IllegalArgumentException e)
			{
				if(e instanceof NumberFormatException) throw new ParserException("Invalid arguments."); // Cannot be casted because the user fucked up
				else throw new ParserException("Parsing error, report to dev.", e); // Cannot be casted because the dev fucked up
			}
		}

		return outputList;
	}

	private static Pattern usernamePattern = Pattern.compile("^.*#\\d{4}$");
	private static Pattern idPattern = Pattern.compile("^\\d{18}$");

	/**
	 * Binds an username or an UID to an actual {@code IUser}.
	 * The following username input formats are allowed:
	 * <ul>
	 *     <li>18-digit UID: "123456789012345678"</li>
	 *     <li>Username and 4-digit discriminator: "Username#1234"</li>
	 * </ul>
	 *
	 * Unknown formats will cause a {@link ParserException}.
	 *
	 * @param input String to parse
	 * @param client Discord instance to look for actual users
	 * @return A filled Optional when an user was found, an empty one otherwise.
	 * @throws ParserException Unknown format
	 */
	public static Optional<IUser> checkParseUsernameOrID(String input, IDiscordClient client) throws ParserException
	{
		IUser foundUser = null;

		if(idPattern.matcher(input).matches()) foundUser = client.getUserByID(input); // UID
		else if(usernamePattern.matcher(input).matches()) // username#1234
		{
			for(IUser iUser : client.getUsers()) // Greedy
				if(input.equalsIgnoreCase(iUser.getName() + "#" + iUser.getDiscriminator()))
				{
					foundUser = iUser;
					break; // Limit greediness
				}
		}
		else throw new ParserException("Invalid argument."); // ?

		return Optional.ofNullable(foundUser);
	}

	/**
	 * Binds a guild or an ID to an actual {@link IGuild}.
	 * The following guild input formats are allowed:
	 * <ul>
	 *     <li>18-digit UID: "123456789012345678"</li>
	 *     <li>Guild name: "Discord Developers"</li>
	 * </ul>
	 *
	 * @param input String to parse
	 * @param client Discord instance to look for guilds
	 * @return A filled Optional when a guild was found, an empty one otherwise.
	 */
	public static Optional<IGuild> checkParseGuildOrID(String input, IDiscordClient client)
	{
		IGuild foundGuild = null;

		if(idPattern.matcher(input).matches()) foundGuild = client.getGuildByID(input); // ID
		else
			for(IGuild i : client.getGuilds()) // Name
				if(input.equalsIgnoreCase(i.getName()))
				{
					foundGuild = i;
					break;
				}

		return Optional.ofNullable(foundGuild);
	}

	/**
	 * Thrown whenever any parse operation failed.
	 */
	public static class ParserException extends RuntimeException
	{
		public ParserException(String message, Throwable cause)
		{
			super(message, cause);
		}

		public ParserException(Throwable cause)
		{
			super(cause);
		}

		public ParserException()
		{
			super();
		}

		public ParserException(String message)
		{
			super(message);
		}
	}
}
